# -*- coding: utf-8 -*-
"""
Created on Mon Mar  4 11:32:01 2019

@author: yoelr
"""

from .. import Unit
from .decorators import cost, design

@cost('Flow rate', CE=576, cost=533471, S=333333, exp=0.6)
@design('Flow rate', 'kg/hr', lambda self: self._ins[0].massnet)
class MagneticSeparator(Unit):
    _N_outs = 1
    _linkedstreams = True
    


# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 22:07:01 2018

@author: yoelr
"""
from ._tank import MixTank
from .decorators import attach_heat_exchanger

@attach_heat_exchanger
class EnzymeTreatment(MixTank):
    """Create an EnzymeTreatment unit that is cost as a MixTank with a heat exchanger."""
    _N_outs = 1
    _kwargs = {'T': 298.15}  # operating temperature (K)
    
    #: Residence time (hr)
    _tau = 1
    
    def _setup(self):
        self.outs[0].T = self._kwargs['T']
    
    def _run(self):
        feed = self.ins[0]
        out = self.outs[0]
        out._mol[:] = self._mol_in
        out.phase = feed.phase
        out.P = feed.P
    


# -*- coding: utf-8 -*-
"""
Created on Wed Jun  5 21:06:29 2019

@author: yoelr
"""
from ... import Unit
from copy import copy
from math import ceil
from pandas import DataFrame
from ._extend import extend_summary
__all__ =('N',)

_index = ('Basis', 'Limit')

def _N(self):
    table = self.N_options
    Design = self._results['Design']
    for i in table:
        basis, ub = table[i]
        value = Design[basis]
        Design[i] = N = ceil(value/ub)
        Design[basis] = value/N

def N(basis, name='N', *, limit):
    """Add design item for number of parallel unit items based on basis limit.
    
    **Parameters**
    
        **basis:** Name of Design item used for scaling.
        
        **name:** Name of item.
        
        **limit:** Maximum value of basis.
        
    **Examples**
    
        :doc:`Unit decorators`
    
    """
    def N_decorator(cls):
        extend_summary(cls)
        if '_N' in cls.__dict__:
            raise RuntimeError("cannot decorate class as '_N' method is already implemented")
        if hasattr(cls, 'N_options'):
            if 'N_options' not in cls.__dict__:
                cls.N_options = copy(cls.N_options)
            if name in cls.N_options:
                raise ValueError("name '{name_}' not available, must pass a name not previously used")
            cls.N_options[name] = (basis, limit)
        else:
            cls.N_options = DataFrame([basis, limit],
                                      columns=(name,),
                                      index=_index)
            cls._N = _N
        return cls
    return N_decorator


# -*- coding: utf-8 -*-
"""
Created on Wed May  1 19:05:53 2019

@author: yoelr
"""
import pandas as pd
from copy import copy
from ._extend import extend_finalize

__all__ = ('cost',)

_index = pd.MultiIndex.from_tuples(
        (('Basis', 'param'),
         ('Basis', 'units'),
         ('Basis', 'value'),
         ('Number', ''),
         ('CEPCI', ''),
         ('Cost (USD)', ''),
         ('Exponent', ''),
         ('Electricity (kW)', '')))

def _cost(self):
    table = self.cost_options
    Cost = self._results['Cost']
    Design = self._results['Design']
    check = isinstance
    net_e = 0
    for i in table:
        param, units, value, N, CEPCI, cost, n, e = table[i]
        if check(N, str): N = Design[N]
        S = Design[param]/value
        Cost[i] = N*self.CEPCI/CEPCI*cost*S**n
        net_e += (e*N*S)
    if net_e: self._power_utility(net_e)

def cost(basis, name=None, *, cost, exp, CE, S=1, kW=0, N=1):    
    r"""Add item purchase cost based on exponential scale up:
    
    :math:`C_{f.o.b.} = (N)(CE)(cost)(\frac{basis}{S})^{exp}` 
    
    :math:`Electricity\ rate = (N)(kW)(\frac{basis}{S})`
    
    **Parameters**
    
        **basis:** Name of Design item used for scaling.
        
        **name:** Name of item.
        
        **cost:** Cost of item.
        
        **exp:** Exponential factor.
        
        **CE:** Chemical engineering plant cost index.
        
        **S:** Size.
        
        **kW:** Electricity rate.
        
        **N:** Number of items or name of Design item that represents N.
        
    **Examples**
    
        :doc:`Unit decorators`
    
    """
    def cost_decorator(cls):
        extend_finalize(cls)
        if kW: cls._has_power_utility = True
        try:
            units = cls._units[basis]
        except KeyError:
            raise RuntimeError(f'units of cost basis ({basis}) is not available in "{cls.__name__}._units" dictionary')
        data = [basis, units, S, N, CE, cost, exp, kW]
        if hasattr(cls, 'cost_options'):
            if 'cost_options' not in cls.__dict__:
                cls.cost_options = copy(cls.cost_options)
            if not name:
                raise ValueError("must pass a 'name' for purchase cost item")
            if name in cls.cost_options:
                raise ValueError("name '{name_}' not available, must pass a name not previously used")
            cls.cost_options[name or cls.line] = data
        else:
            if '_cost' in cls.__dict__:
                raise RuntimeError("cannot decorate class as '_cost' method is already implemented")
            columns = pd.Index((name or cls.line,), name='Item')
            cls.cost_options = pd.DataFrame(data, _index, columns)
            cls._cost = _cost
        return cls
    return cost_decorator




# -*- coding: utf-8 -*-
"""
Created on Sat Aug 18 13:42:33 2018

@author: yoelr
"""
from . import compounds as comps
import numpy as np

Chemical = comps.Chemical
Compound = comps.Compound
inf = np.inf
__all__ = ('Species',)

# TODO: Fix how Dortmund groups are selected in thermo.Chemical. Glycerol defaults to wrong value.

# %% Managing ChEDL thermodynamic properties
setattr_ = object.__setattr__

class Species:
    """Create Species object that contains Compound objects as attributes.

    **Parameters**

        ***IDs:** [str] Strings should be one of the following [-]:
                  * Name, in IUPAC form or common form or a synonym registered in PubChem
                  * InChI name, prefixed by 'InChI=1S/' or 'InChI=1/'
                  * InChI key, prefixed by 'InChIKey='
                  * PubChem CID, prefixed by 'PubChem='
                  * SMILES (prefix with 'SMILES=' to ensure smiles parsing)
                  * CAS number
         
        **Optional**
        
        **cls:** [type Compound] The type of objects that will be created and stored as attributes. Defaults to biosteam.Chemical
        
    """
    units = Compound.units
    _immutable = False
    _compounds = _IDs = _CAS = _Nspecies = None
    
    @classmethod
    def tospecies(cls, compounds, copy=True):
        """Return a Species object from an iterable of Compound objects."""
        if not copy and isinstance(compounds, Species): return compounds
        self = cls.__new__(cls)
        for c in compounds: 
            if isinstance(c, Compound): setattr_(self, c.ID, c)
        return self
    
    def __init__(self, *IDs, cls=None):
        # Sphinx docs look ugly if a class is in the default argument
        if not IDs: return
        if cls is None: cls = Chemical
        
        # Set Compound object attributes
        for n in IDs:
            try:
                compound = cls(n)
            except:
                raise Exception(f"Compound object, '{n}', not defined in data bank")
            setattr_(self, n, compound)

    def _read_only(self):
        if self._immutable: return
        tup = tuple
        compounds = tup(self)
        IDs = tup(i.ID for i in compounds)
        CAS = tup(i.CAS for i in compounds)
        Nspecies = len(IDs)
        index = tup(range(Nspecies))
        me = self.__dict__
        me['_immutable'] = True
        me['_num_compounds'] = tup(zip(index, compounds))
        me['_num_IDs'] = tup(zip(index, IDs))
        me['_compounds'] = compounds
        me['_IDs'] = IDs
        me['_CAS'] = CAS
        me['_Nspecies'] = Nspecies
        me['_MW'] = np.array([i.MW for i in compounds])
        
    def __setattr__(self, ID, compound):
        if self._immutable:
            raise ValueError('cannot alter species object attached to Stream objects')
        elif isinstance(compound, Compound):
            compound.ID = ID
            setattr_(self, ID, compound)
        elif ID == 'IDs':
            setattr_(self, '__dict__', {i: getattr(self, i) for i in compound})
        elif ID == 'CAS':
            index = compound.index if hasattr(compound, 'index') else tuple(compound).index
            setattr_(self, '__dict__', {i.ID: i for i in sorted(self, key=lambda x: index(x.CAS))})
        else:
            raise TypeError('can only set Compound objects as attributes')
    
    def __delattr__(self, ID):
        if self._immutable:
            raise ValueError('cannot alter species object attached to Stream objects')
        else:
            super().__delattr__(ID)
    
    def __len__(self):
        return self._Nspecies or len(self.__dict__)
    
    def __iter__(self):
        return iter(self._compounds or self.__dict__.values())
    
    @property
    def CAS(self):
        return self._CAS or tuple(s.CAS for s in self)
    
    @property
    def IDs(self):
        return self._IDs or tuple(self.__dict__.keys())

    def getprops(self, species_IDs, prop_ID, T, P, phase):
        """Return list of the desired property, prop_ID, for each compound in species_ID."""
        props = []; gat = getattr
        for sID in species_IDs:
            s = gat(self, sID)
            s.T = T
            s.P = P
            s.phase = phase
            props.append(gat(s, prop_ID))
        return props

    def setprops(self, species_IDs, prop_ID, new_values):
        """Set new values to specie property, prop_ID, for given species_IDs."""
        getattr_ = getattr
        for ID in species_IDs: 
            setattr_(getattr_(self, ID), prop_ID, new_values)

    ### Material Properties ###

    # General methods for getting ideal mixture properties
    def _props(self, ID, mol, T, P, phase):
        """Return component property list."""
        out = np.zeros(self._Nspecies)
        getattr_ = getattr
        ic = self._num_compounds.__iter__().__next__
        for m in mol:
            if m: 
                i, c = ic()
                c.P = P
                c.T = T
                c.phase = phase
                prop = getattr_(c, ID)
                out[i] = prop if prop else 0
            else: ic()
        return out

    def _propflows(self, ID, mol, T, P, phase):
        """Return array of component properties * kmol/hr."""
        return self._props(ID, mol, T, P, phase) * mol

    def _propflow(self, ID, mol, T, P, phase):
        """Return flow of component properties."""
        return (self._props(ID, mol, T, P, phase)*mol).sum()

    def _prop(self, ID, mol, T, P, phase):
        """Return molar weighted average property."""
        return (self._props(ID, mol, T, P, phase)*mol).sum().sum()

    def _mixedpropflows(self, ID, molarray, T, P):
        """Return array of component properties * kmol/hr."""
        pf = self._propflows
        return (pf(ID, molarray[0], T, P, 's')
              + pf(ID, molarray[1], T, P, 'l')
              + pf(ID, molarray[2], T, P, 'l')
              + pf(ID, molarray[3], T, P, 'g'))
    
    def _mixedpropflow(self, ID, molarray, T, P):
        """Return flow of component properties."""
        return self._mixedpropflows(ID, molarray, T, P).sum()
    
    def _mixedprop(self, ID, molarray, T, P):
        """Return molar weighted average property."""
        return self._mixedpropflows(ID, molarray, T, P).sum()/molarray.sum()  
    
    ### Equilibrium ###
    
    def _equilibrium_species(self, mol):
        """Return species and indexes of species in equilibrium."""
        species = []; index = []; ic = self._num_compounds.__iter__().__next__
        for m in mol:
            if m:
                i, c = ic()
                if c.UNIFAC_Dortmund_groups and c.Tb not in (inf, None):
                    species.append(c); index.append(i)
            else: ic()
        return species, index

    def _heavy_species(self, mol):
        """Return species and indexes of heavy species not in equilibrium."""
        species = []; index = []; ic = self._num_compounds.__iter__().__next__
        for m in mol:
            if m:
                i, c = ic()
                if c.Tb in (inf, None) or not c.UNIFAC_Dortmund_groups:
                    species.append(c); index.append(i)
            else: ic()
        return species, index

    def _light_species(self, mol):
        """Return species and indexes of light species not in equilibrium."""
        species = []; index = []; ic = self._num_compounds.__iter__().__next__
        for m in mol:
            if m:
                i, c = ic()
                if c.Tb is -inf:
                    species.append(c); index.append(i)
            else: ic()
        return species, index
    
    # Other
    def _reorder(self, values, species_IDs):
        """Return a reordered array with zeros in place of missing species IDs."""
        IDs = self._IDs
        index = IDs.index
        array = np.zeros_like(IDs, dtype=float)
        array[[index(i) for i in species_IDs]] = values
        return array
    
    # Representation
    def __repr__(self):
        return '<Species: ' + str(self.IDs).replace("'", '')[1:-1] + '>'

    



# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 22:14:01 2018

@author: yoelr
"""
from .. import Unit
from .decorators import cost, design
from .metaclasses import splitter, run_split_with_mixing

@cost('Flow rate', cost=1.5e6, CE=541.7, exp=0.6, S=335e3, kW=2010)
@design('Flow rate', 'kg/hr', lambda self: self._ins[0].massnet)
class CrushingMill(Unit, metaclass=splitter):
    _run = run_split_with_mixing
    



# -*- coding: utf-8 -*-
"""
Created on Mon Apr 29 18:28:55 2019

@author: yoelr
"""
from ... import Unit
from ..decorators import cost, design

__all__ = ('CoolingTower', 'CoolingTowerWithPowerDemand')

@cost('Flow rate', cost=1100, CE=567, exp=0.68)
@design('Flow rate', 'gpm', lambda self: 0.07934*self.ins[0].molnet)
class CoolingTower(Unit):
    """Create a cooling tower that is cost based on flow rate of cooling water.
    
    **ins**
    
        [0] Cooling water
        
    **outs**
    
        [0] Cooling water
    
    """
    _N_ins = 1; _N_outs = 1
    _linkedstreams = True

   
class CoolingTowerWithPowerDemand(CoolingTower):
    _has_power_utility = True
    _N_heat_utilities = 1
    power_consumption = 0.03 #: kWhr per kmol cooling water
    def _cost(self):
        super()._cost()
        q = self.ins[0].molnet # kmol/hr
        self._power_utility(q*self.power_consumption)
        hu = self._heat_utilities[0]
        cw = hu.cooling_agents['Cooling water']
        hu.ID = 'Cooling water'
        hu.flow = -q
        hu.cost = -q*cw['Price (USD/kmol)']


# -*- coding: utf-8 -*-
"""
Created on Fri Jun  7 00:53:29 2019

extend_summary extends the Unit with the following abstract methods:

**_N()**
    Add number of units to "Design" dictionary (called after `_design`).

extend_finalize extends the Unit with the following abstract methods:

**_spec()**
    Calculate specification cost factors and update purchase prices (called after `_cost`).
    
**_end():**
    Finish setting purchase prices and utility costs (called after `_spec`).


@author: yoelr
"""
from ... import Unit

__all__ = ('extend_finalize', 'extend_summary')

def extend_summary(cls):
    if cls._summary is Unit._summary:
            cls._summary = _summary
    elif cls._summary is not _summary:
        raise RuntimeError("cannot decorate Unit class with an implemented '_summary' method")

def extend_finalize(cls):
    if cls._finalize is Unit._finalize:
        cls._finalize = _finalize
    elif cls._finalize is not _finalize:
        raise RuntimeError("cannot decorate Unit class with an implemented '_summary' method")

def _summary(self):
    """Calculate all results from unit run."""
    self._design()
    self._N()
    self._finalize()

def _finalize(self):
    """Run all cost methods and finalize purchase and utility cost."""
    self._cost()
    self._spec()
    self._end()
    self._update_utility_cost()
    self._update_purchase_cost()


# -*- coding: utf-8 -*-
"""
Created on Mon Mar  4 10:57:12 2019

@author: yoelr
"""
from .. import Unit
from .decorators import cost, design

@cost('Flow rate', cost=2.5e6, CE=567.3, exp=0.6, S=500e3, kW=3000)
@design('Flow rate', 'kg/hr', lambda self: self._ins[0].massnet)
class Shredder(Unit):
    _linkedstreams = True
    _N_outs = 1


# -*- coding: utf-8 -*-
"""
Created on Mon May  6 17:19:41 2019

@author: yoelr
"""

__all__ = ('design',)

def _design(self):
    D = self._results['Design']
    for i, j in self._design_basis: D[i] = j(self)

def design(name, units, func):    
    """Add design item.
    
    **Parameters**
    
        **name:** Name of design item.
        
        **units:** Units of measure of design item.
        
        **func:** Should return design item given the Unit object.
        
    **Examples**
    
        :doc:`Unit decorators`
    
    """
    def design_decorator(cls):
        if not cls._units:
            cls._units = {}
        elif '_units' not in cls.__dict__:
            cls._units = cls._units.copy()
        cls._units[name] = units
        if hasattr(cls, '_design_basis'):
            cls._design_basis.append((name, func))
        else:
            cls._design_basis = [(name, func)]
            cls._design = _design
        return cls
    return design_decorator




# -*- coding: utf-8 -*-
"""
Created on Thu Mar  7 23:17:38 2019

@author: yoelr
"""
import numpy as np
import pandas as pd
from ._state import State

__all__ = ('Model',)

# %% Functions

paramindex = lambda params: pd.MultiIndex.from_arrays(
                                ([f.element_name for f in params], 
                                 [f.name for f in params]),
                                names=('Element', 'Parameter'))

    
# %% Grid of simulation blocks

class Model(State):
    """Create a Model object that allows for evaluation over a sample space.
    
    **Parameters**
    
        **ID:** [str] ID of metric        
    
        **system:** [System] Should reflect the model state.
    
        **metric:** [function] Should return metric value.
    
    **Examples**

         :doc:`Advanced simulation`
    
    """
    __slots__ = ('_ID',      # [str] Should be the metric name.
                 '_table',   # [DataFrame] All arguments and results.
                 '_metric',  # [function] Should return metric being evaluated.
                 '_index',   # list[int] Order of sample evaluation for performance.
                 '_samples') # [array] Argument sample space.
    
    def __init__(self, ID, system, metric):
        super().__init__(system)
        self._ID = ID
        self._metric = metric
        self._samples = self._table = None
    
    def _erase(self):
        """Erase cached data."""
        self._model = self._table = self._samples = None
    
    @property
    def table(self):
        """[DataFrame] Table of the sample space with results in the final column."""
        return self._table
    
    def load_samples(self, samples):
        """Load samples for evaluation"""
        if not self._model: self._loadmodel()
        params = self._params
        paramlen = len(params)
        if not isinstance(samples, np.ndarray):
            raise TypeError(f'samples must be an ndarray, not a {type(samples).__name__} object')
        elif samples.ndim != 2:
            raise ValueError('samples must be 2 dimensional')
        elif samples.shape[1] != paramlen:
            raise ValueError(f'number of parameters in samples ({samples.shape[1]}) must be equal to the number of parameters ({len(params)})')
        key = lambda x: samples[x][i]
        index = list(range(len(samples)))
        for i in range(paramlen-1,  -1, -1):
            if not params[i].system: break
            index.sort(key=key)
        self._index = index
        self._table = pd.DataFrame(samples, columns=paramindex(params))
        self._table[self._ID] = None
        self._samples = samples
        
    def evaluate(self, default=None):
        """Evaluate metric over the argument sample space and save values to `table`."""
        if not self._model: self._loadmodel()
        # Setup before simulation
        metric = self._metric
        values = []
        add = values.append
        model = self._model
        samples = self._samples
        if samples is None: raise ValueError('must load samples or distribution before evaluating')
        for i in self._index: 
            try: 
                model(samples[i])
                add(metric())
            except: add(default)
        self.table[self._ID] = values
    
    def __call__(self, sample):
        super().__call__(sample)
        return self._metric()
    
    def _repr(self):
        return f'{type(self).__name__}: {self._ID}'
        
   
    
        


# -*- coding: utf-8 -*-
"""
Created on Mon May 20 22:59:57 2019

@author: yoelr
"""
__all__ = []

from ._splitter import *
from ._final import *
from . import _splitter
from . import _final
__all__.extend(_splitter.__all__)
__all__.extend(_final.__all__)


# -*- coding: utf-8 -*-
"""
Created on Thu Jun  6 00:55:03 2019

Mechanical equipment

@author: yoelr
"""

from ... import Unit, Mixer, Stream
from ..metaclasses import splitter
from .. import decorators
from .designtools import CEPCI_by_year as CE

# %% Decorators 

_massflow_units = {'Flow rate': 'kg/hr'}
def _design_kg_hr(self):
    self._results['Design']['Flow rate'] = self._ins[0].massnet

def cost_kg_hr(*, cost, exp, S, kW=0, CE=CE[2009], N=1, basis='Flow rate'):
    def decorator(cls):
        cls._design = _design_kg_hr
        cls._units = _massflow_units
        return decorators.cost(basis, cost=cost, exp=exp,
                               CE=CE, S=S, kW=kW, N=N)(cls)
    return decorator

# %% Units

@cost_kg_hr(cost=13329690, exp=0.6, S=94697, kW=511.321)
class FeedStockHandling(Unit): BM = 1.7
    
@cost_kg_hr(cost=6000, exp=0.5, S=136260)
class SulfuricAcidMixer(Unit): BM = 1.0
    
@cost_kg_hr(cost=19812400, exp=0.6, S=83333, kW=5290)
class PretreatmentReactorSystem(Unit):
    BM = 1.5
    def _init(self):
        self._water_mass = Stream.indices('Water')
    def _design(self):
        feed = self._ins[0]
        self._results['Design']['Flow rate'] = feed.massnet - feed.mass[self._water_index]
        
    


# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 22:49:58 2018

@author: yoelr
"""
import numpy as np
from .. import Unit
from .decorators import design, cost

@cost('Volume', CE=525.4, cost=15000, exp=0.55, kW=1.5)
@design('Volume', 'm^3', lambda self: self._tau*self._volnet_out/0.8)
class Transesterification(Unit):
    """Create a transesterification reactor that converts 'Lipid' and 'Methanol' to 'Biodiesel' and 'Glycerol'. Finds the amount of catalyist 'NaOCH3' required and consumes it to 'NaOH' and 'Methanol'.
    
    **Parameters**
    
        **efficiency:** Efficiency of conversion (on a 'Lipid' basis)
        
        **r:** Methanol to lipid molar ratio
        
        **catalyst_molfrac:** Molar fraction of catalyst in methanol feed 
        
        **T:** Operating temperature (K)
    
    **ins**
    
        [0] Lipid feed
        
        [1] Fresh methanol and catalyst
        
    **outs**
    
        [0] Product
    
    **Examples**
    
        :doc:`Transesterification Example`
    
    """
    _kwargs = {'efficiency': None,  # fraction of theoretical conversion
               'r': None,  # Methanol to lipid molar ratio
               'T': None,
               'catalyst_molfrac': None}  # operating temperature (K)    
    _bounds = {'Volume': (0.1, 20)}
    _tau = 1
    _N_ins = 2
    _N_outs = 1
    _N_heat_utilities = 1

    def _setup(self):
        self._index = self.outs[0].indices('Lipid', 'Methanol', 'Glycerol', 'Biodiesel', 'NaOH', 'NaOCH3')

    def _run(self):
        feed, fresh_Methanol = self.ins
        out = self.outs[0]
        eff, r, T, catalyst_molfrac = (self._kwargs[i] for i in (
            'efficiency', 'r', 'T', 'catalyst_molfrac'))

        # Reactant positions
        lipid_pos, Methanol_pos, Glycerol_pos, biodiesel_pos, NaOH_pos, NaOCH3_pos = self._index
        lipid = feed.mol[lipid_pos]

        # Reaction conversions by mol (to add to the feed)
        lipid_ch = -eff*lipid  # - free_lipid_ch/3
        bd_ch = eff*lipid*3
        gly_ch = -lipid_ch
        dummy = r*lipid
        NaOCH3_in = dummy*catalyst_molfrac  # from methanol stream
        Methanol_in = dummy - NaOCH3_in  # from methanol stream
        Methanol_ch = dummy + lipid_ch
        NaOCH3_ch = 0
        NaOH_ch = NaOCH3_in

        change_mol = np.array((lipid_ch, Methanol_ch, gly_ch,
                               bd_ch, NaOCH3_ch, NaOH_ch))

        # Output stream
        fresh_Methanol.mol[[Methanol_pos, NaOCH3_pos]] = (Methanol_in, NaOCH3_in)
        index = [lipid_pos, Methanol_pos, Glycerol_pos,
                biodiesel_pos, NaOCH3_pos, NaOH_pos]
        out.phase = 'l'
        out.mol[index] = feed.mol[index] + change_mol
        out.T = T
        out.P = feed.P

    def _end(self):
        self._heat_utilities[0](self._Hnet, self.outs[0].T)
        
    
        
        
